import logging
import math

from tusclient.client import TusClient
from progressbar import progressbar

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def upload_file(path, host):
    tus_url = host + '/files/'
    logger.info(f"Uploading file {path}...")
    tus = TusClient(tus_url)
    chunk_size = 1*1024*1024 # 1 Mb
    uploader = tus.uploader(path, chunk_size=chunk_size)
    num_chunks = math.ceil(uploader.get_file_size()/chunk_size)

    for _ in progressbar(range(num_chunks)):
        uploader.upload_chunk()
    return uploader.url




#!/usr/bin/env python

import argparse
import os
from uuid import uuid1
import logging

from progressbar import progressbar

from ..util import md5sum

from .create_media import create_media
from .determine_transcode import determine_transcode
from .transcode import convert_streaming
from .transcode import convert_archival
from .transcode import convert_audio
from .make_thumbnails import make_thumbnails
from .upload_transcoded_video import upload_transcoded_video

logger = logging.getLogger(__name__)

def parse_args():
    parser = argparse.ArgumentParser(description='Full transcode pipeline on a directory of files.')
    parser.add_argument('path', type=str, help='Path to directory containing video files, or a video file.')
    parser.add_argument('--extension', type=str, default='mp4', help='File extension to upload. '
                                                                     'Ignored if path is a file.')
    parser.add_argument('--host', type=str, default='https://www.tatorapp.com', help='Host URL.')
    parser.add_argument('--token', type=str, help='REST API token.')
    parser.add_argument('--project', type=int, help='Unique integer specifying project ID.')
    parser.add_argument('--type', type=int, help='Unique integer specifying a media type.')
    parser.add_argument('--section', type=str, help='Media section name.')
    return parser.parse_args()

def get_file_paths(path):
    base, _ = os.path.splitext(path)
    paths = {
        'original': path,
        'transcoded': base + '_transcoded',
        'thumbnail': base + '_thumbnail.jpg',
        'thumbnail_gif': base + '_thumbnail_gif.gif',
        'segments': base + '_segments.json',
    }
    return paths

def transcode_single(path, args, gid):
    """Transcodes a single file.
    """
    paths = get_file_paths(path)

    # Get md5 for the file.
    md5 = md5sum(paths['original'])

    # Get base filename.
    name = os.path.basename(paths['original'])

    # Create the media object.
    media_id = create_media(args.host, args.token, args.project, args.type, args.section, name, md5)

    # Make thumbnails.
    make_thumbnails(args.host, args.token, media_id, paths['original'], paths['thumbnail'],
                    paths['thumbnail_gif'])

    # Determine transcodes that need to be done.
    workloads = determine_transcode(args.host, args.token, args.type, path, group_to=1080)

    # Transcode the video file.
    for workload in workloads:
        category = workload['category']
        del workload['category']
        if category == 'streaming':
            convert_streaming(**workload, host=args.host, token=args.token, media=media_id,
                              outpath=paths['transcoded'], gid=gid, uid=str(uuid1()))
        elif category == 'archival':
            del workload['resolutions']
            convert_archival(**workload, host=args.host, token=args.token, media=media_id,
                             outpath=paths['transcoded'])
        elif category == 'audio':
            del workload['resolutions']
            del workload['raw_width']
            del workload['raw_height']
            convert_audio(**workload, host=args.host, token=args.token, media=media_id,
                          outpath=paths['transcoded'])
    
if __name__ == '__main__':
    args = parse_args()
    gid = str(uuid1())
    if os.path.isdir(args.path):
        file_list = []
        for root, dirs, files in os.walk(args.path):
            for fname in files:
                if fname.endswith(args.extension):
                    path = os.path.join(root, fname)
                    file_list.append(path)
        for path in progressbar(file_list):
            transcode_single(path, args, gid)
    else:
        transcode_single(args.path, args, gid)



import tator
from uuid import uuid1

def test_transcode_fault_handling(host, token, project):
    """ Verify Error Handling occurs for transcode endpoint """
    tator_api = tator.get_api(host, token)
    exception = False
    try:
        video_obj = tator_api.transcode(project,
                                        transcode_spec={'gid':str(uuid1()),
                                                        'md5':"1234",
                                                        'name':"blah",
                                                        'section':"blah",
                                                        'type':200,
                                                        'uid':str(uuid1()),
                                                        'url':"blah"})
    except Exception as e:
        exception = True
        code = e.status

    assert(exception)
    assert(code == 400)


    




import datetime
import random
import uuid
import time

import tator
from ._common import assert_close_enough

def random_localization(project, box_type, video_obj, post=False):
    x = random.uniform(0.0, 1.0)
    y = random.uniform(0.0, 1.0)
    w = random.uniform(0.0, 1.0 - x)
    h = random.uniform(0.0, 1.0 - y)
    attributes = {
        'test_bool': random.choice([False, True]),
        'test_int': random.randint(-1000, 1000),
        'test_float': random.uniform(-1000.0, 1000.0),
        'test_enum': random.choice(['a', 'b', 'c']),
        'test_string': str(uuid.uuid1()),
        'test_datetime': datetime.datetime.now().isoformat(),
        'test_geopos': [random.uniform(-180.0, 180.0), random.uniform(-90.0, 90.0)],
    }
    out = {
        'x': x,
        'y': y,
        'width': w,
        'height': h,
        'project': project,
        'type': box_type,
        'media_id': video_obj.id,
        'frame': random.randint(0, video_obj.num_frames - 1),
    }
    if post:
        out = {**out, **attributes}
    else:
        out['attributes'] = attributes
    return out

def test_localization_crud(host, token, project, video_type, video, box_type):
    tator_api = tator.get_api(host, token)
    video_obj = tator_api.get_media(video)

    # These fields will not be checked for object equivalence after patch.
    exclude = ['project', 'type', 'media_id', 'id', 'meta', 'user']

    # Test bulk create.
    num_localizations = random.randint(2000, 10000)
    boxes = [
        random_localization(project, box_type, video_obj, post=True)
        for _ in range(num_localizations)
    ]
    box_ids = []
    for response in tator.util.chunked_create(tator_api.create_localization_list,
                                         project, localization_spec=boxes):
        box_ids += response.id
    assert len(box_ids) == len(boxes)
    print(f"Created {len(box_ids)} boxes!")

    # Test single create.
    box = random_localization(project, box_type, video_obj, post=True)
    response = tator_api.create_localization_list(project, localization_spec=[box])
    assert isinstance(response, tator.models.CreateListResponse)
    box_id = response.id[0]

    # Patch single box.
    patch = random_localization(project, box_type, video_obj)
    response = tator_api.update_localization(box_id, localization_update=patch)
    assert isinstance(response, tator.models.MessageResponse)
    print(response.message)

    # Get single box.
    updated_box = tator_api.get_localization(box_id)
    assert_close_enough(patch, updated_box, exclude)
    
    # Delete single box.
    response = tator_api.delete_localization(box_id)
    assert isinstance(response, tator.models.MessageResponse)
    print(response.message)

    # ES can be slow at indexing so wait for a bit.
    time.sleep(5)

    # Bulk update box attributes.
    bulk_patch = random_localization(project, box_type, video_obj)
    bulk_patch = {'attributes': bulk_patch['attributes']}
    params = {'media_id': [video], 'type': box_type}
    response = tator_api.update_localization_list(project, **params,
                                                  attribute_bulk_update=bulk_patch)
    assert isinstance(response, tator.models.MessageResponse)
    print(response.message)

    # Verify all boxes have been updated.
    boxes = tator_api.get_localization_list(project, **params)
    dataframe = tator.util.to_dataframe(boxes)
    assert(len(boxes)==len(dataframe))
    for box in boxes:
        assert_close_enough(bulk_patch, box, exclude)
    
    # Delete all boxes.
    response = tator_api.delete_localization_list(project, **params)
    assert isinstance(response, tator.models.MessageResponse)
    time.sleep(1)

    # Verify all boxes are gone.
    boxes = tator_api.get_localization_list(project, **params)
    assert boxes == []



#!/usr/bin/env python

import argparse
import logging

from ..util.get_api import get_api
from ..util.get_parser import get_parser
from ..openapi.tator_openapi.models import MessageResponse

logger = logging.getLogger(__name__)

def parse_args():
    parser = get_parser()
    parser.add_argument('--project', type=int, help='Unique integer specifying project ID.')
    parser.add_argument('--job_type', type=str, help='One of upload, download, algorithm.')
    parser.add_argument('--gid', type=str, help='A UUID generated for the group.')
    parser.add_argument('--uid', type=str, help='A UUID generated for the job.')
    parser.add_argument('--state', type=str, help='One of queued, failed or started.')
    parser.add_argument('--message', type=str, help='Progress message.')
    parser.add_argument('--progress', type=int, help='Progress (0-100).')
    parser.add_argument('--name', type=str, help='Name of the file.')
    parser.add_argument('--section', type=str, help='Media section, if this is for an upload.')
    parser.add_argument('--sections', type=str, help='Comma-separated list of media sections, if this is for an algorithm.')
    parser.add_argument('--media_ids', type=str, help='Comma-separated list of media IDs, if this is for an algorithm.')
    return parser.parse_args()

def send_progress(args):
    api = get_api(host=args.host, token=args.token)
    aux = {}
    if args.section:
        aux['section'] = args.section
    if args.sections:
        aux['sections'] = args.sections
    if args.media_ids:
        aux['media_ids'] = args.media_ids
    
    response = api.progress(args.project, progress_spec=[{
        'job_type': args.job_type,
        'gid': args.gid,
        'uid': args.uid,
        'state': args.state,
        'message': args.message,
        'progress': args.progress,
        'name': args.name,
        **aux,
    }])
    assert isinstance(response, MessageResponse)

if __name__ == '__main__':
    args = parse_args()
    send_progress(args)



import tator
import subprocess

def test_local_transcode(host, token, project, video_type, video_file):
    cmd = [
        'python3', '-m', 'tator.transcode', video_file,
        '--host', host,
        '--token', token,
        '--project', str(project),
        '--type', str(video_type),
        '--section', 'Locally transcoded',
    ]
    subprocess.run(cmd, check=True)



#!/usr/bin/env python

import argparse
import logging
import json
import math
import subprocess
import os

from tusclient.client import TusClient
from progressbar import progressbar
import requests

from .make_fragment_info import make_fragment_info

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def parse_args():
    parser = argparse.ArgumentParser(description='Uploads transcoded video.')
    parser.add_argument('--original_path', type=str, help='Original video file.')
    parser.add_argument('--original_url', type=str, help='Upload URL of original file. If given, upload is skipped.')
    parser.add_argument('--transcoded_path', type=str, help='Transcoded video directory (contains multiple video resolutions).')
    parser.add_argument('--thumbnail_path', type=str, help='Thumbnail file.')
    parser.add_argument('--thumbnail_gif_path', type=str, help='Thumbnail gif file.')
    parser.add_argument('--tus_url', type=str, default='https://www.tatorapp.com/files/', help='TUS URL.')
    parser.add_argument('--url', type=str, default='https://www.tatorapp.com/rest', help='REST API URL.')
    parser.add_argument('--token', type=str, help='REST API token.')
    parser.add_argument('--project', type=int, help='Unique integer specifying project ID.')
    parser.add_argument('--type', type=int, help='Unique integer specifying a media type.')
    parser.add_argument('--gid', type=str, help='A UUID generated for the upload group.')
    parser.add_argument('--uid', type=str, help='A UUID generated for the upload.')
    parser.add_argument('--section', type=str, help='Media section name.')
    parser.add_argument('--name', type=str, help='Name of the file.')
    parser.add_argument('--md5', type=str, help='MD5 sum of the media file.')
    parser.add_argument('--progressName', type=str, help='Name to use for progress update')
    return parser.parse_args()

def upload_file(path, tus_url):
    if not tus_url.endswith('/'):
        tus_url += '/'
    logger.info(f"Uploading file {path}...")
    tus = TusClient(tus_url)
    chunk_size = 1*1024*1024 # 1 Mb
    uploader = tus.uploader(path, chunk_size=chunk_size)
    num_chunks = math.ceil(uploader.get_file_size()/chunk_size)

    for _ in progressbar(range(num_chunks)):
        uploader.upload_chunk()
    return uploader.url

def get_metadata(path):
    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream",
        "-print_format", "json",
        "-select_streams", "v",
        "{}".format(path)
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout

    logger.info("Got info = {}".format(output))
    video_info = json.loads(output)
    stream = video_info["streams"][0]
    seconds = float(stream["duration"])

    # Fill in object information based on probe
    codec = stream["codec_name"]
    fps_fractional = stream["avg_frame_rate"].split("/")
    fps = float(fps_fractional[0]) / float(fps_fractional[1])
    if "nb_frames" in stream:
        num_frames = int(stream["nb_frames"])
    else:
        num_frames = round(fps * seconds)
    width = stream["width"]
    height = stream["height"]

    return (codec, fps, num_frames, width, height)

def make_video_definition(disk_file):
    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream",
        "-print_format", "json",
        "-select_streams", "v",
        disk_file,
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout
    video_info = json.loads(output)
    stream_idx=0
    for idx, stream in enumerate(video_info["streams"]):
        if stream["codec_type"] == "video":
            stream_idx=idx
            break
    stream = video_info["streams"][stream_idx]
    video_def = {"resolution": (stream["height"], stream["width"]),
                 "codec": stream["codec_name"],
                 "codec_description": stream["codec_long_name"]}
    return video_def

def make_audio_definition(disk_file):
    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream",
        "-print_format", "json",
        "-select_streams", "a",
        disk_file,
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout
    audio_info = json.loads(output)
    stream_idx=0
    for idx, stream in enumerate(audio_info["streams"]):
        if stream["codec_type"] == "audio":
            stream_idx=idx
            break
    stream = audio_info["streams"][stream_idx]
    audio_def = {"codec": stream["codec_name"],
                 "codec_description": stream["codec_long_name"]}
    return audio_def

def upload_transcoded_video(original_path, transcoded_path,
                            thumbnail_path, thumbnail_gif_path, tus_url, url,
                            token, project, type, gid, uid, section, name, md5,
                            original_url=None, progress_name=None):
    if progress_name is None:
        progress_name = name

    # Upload files
    if original_url is None or original_url == "None":
        logger.info("Uploading original file...")
        original_url = upload_file(original_path, tus_url)
    else:
        logger.info("Skipping original file upload...")

    logger.info("Uploading thumbnail...")
    thumbnail_url = upload_file(thumbnail_path, tus_url)
    logger.info("Uploading thumbnail gif...")
    thumbnail_gif_url = upload_file(thumbnail_gif_path, tus_url)


    media_files={"archival": [],
                 "streaming": []}
    video_def = make_video_definition(original_path)
    video_def['url'] = original_url
    # Get metadata for the original file
    codec, fps, num_frames, width, height = get_metadata(original_path)
    media_files['archival'].append(video_def)

    # Save the video
    out = requests.post(
        f'{url}/SaveVideo/{project}',
        headers={
            "Authorization": f"Token {token}",
            "Content-Type": "application/json",
            "Accept-Encoding": "gzip",
        },
        json={
            'type': type,
            'uid': uid,
            'gid': gid,
            'media_files': media_files,
            'thumbnail_url': thumbnail_url,
            'thumbnail_gif_url': thumbnail_gif_url,
            'name': name,
            'section': section,
            'md5': md5,
            'num_frames': num_frames,
            'fps': fps,
            'codec': codec,
            'width': width,
            'height': height,
            'progressName': progress_name
        },
    )
    out.raise_for_status()

    # We saved the original, now we can proceed to upload
    # the resolutions (one at a time)
    media_id=out.json()['id']

    for root, dirs, files in os.walk(transcoded_path):
        print(f"Processing {files} in {transcoded_path}")
        for vid_file in files:
            media_files={"streaming": []}
            if os.path.splitext(vid_file)[1] == ".m4a":
                # Handle audio file
                audio_path = os.path.join(root, vid_file)
                audio_url = upload_file(audio_path, tus_url)
                audio_def = make_audio_definition(audio_path)
                audio_def["url"] = audio_url
                media_files["audio"] = [audio_def]
            else:
                # Handle video files
                vid_path = os.path.join(root, vid_file)
                base = os.path.splitext(vid_file)[0]
                segments_path = os.path.join(root, f"{base}.json")
                make_fragment_info(vid_path, segments_path)

                logger.info("Uploading transcoded file...")
                transcoded_url = upload_file(vid_path, tus_url)

                logger.info("Uploading segments file...")
                segments_url = upload_file(segments_path, tus_url)

                #Generate video info block
                video_def = make_video_definition(vid_path)
                video_def["url"] = transcoded_url
                video_def["segment_info_url"] = segments_url
                media_files['streaming'].append(video_def)

            out = requests.patch(
            f'{url}/SaveVideo/{project}',
            headers={
                "Authorization": f"Token {token}",
                "Content-Type": "application/json",
                "Accept-Encoding": "gzip",
            },
                json={
                    'media_files': media_files,
                    'id': media_id,
                    'uid': uid,
                    'gid': gid,
                },
            )
            out.raise_for_status()

if __name__ == '__main__':
    args = parse_args()
    upload_transcoded_video(**args)




#!/usr/bin/env python

import argparse
import subprocess
import json
import logging

from ..util import get_api

from .upload import upload_file

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def parse_args():
    parser = argparse.ArgumentParser(description='Makes thumbnails for a video.')
    parser.add_argument('--host', type=str, default='https://www.tatorapp.com', help='Host URL.')
    parser.add_argument('--token', type=str, help='REST API token.')
    parser.add_argument('--media', type=int, help='Unique integer identifying a media.')
    parser.add_argument('input', type=str, help='Path to input file.')
    parser.add_argument("-o", "--output", type=str, help='Path to output thumbnail.')
    parser.add_argument("-g", "--gif", type=str, help='Path to output thumbnail gif.')
    return parser.parse_args()

def get_metadata(path):
    cmd = [
        "ffprobe",
        "-v","error",
        "-show_entries", "stream",
        "-print_format", "json",
        "-select_streams", "v",
        "{}".format(path)
    ]
    output = subprocess.run(cmd, stdout=subprocess.PIPE, check=True).stdout

    logger.info("Got info = {}".format(output))
    video_info = json.loads(output)
    stream = video_info["streams"][0]
    seconds = float(stream["duration"])

    # Fill in object information based on probe
    codec = stream["codec_name"]
    fps_fractional = stream["avg_frame_rate"].split("/")
    fps = float(fps_fractional[0]) / float(fps_fractional[1])
    if "nb_frames" in stream:
        num_frames = int(stream["nb_frames"])
    else:
        num_frames = round(fps * seconds)
    width = stream["width"]
    height = stream["height"]

    return (codec, fps, num_frames, width, height)

def make_thumbnails(host, token, media_id, video_path, thumb_path, thumb_gif_path):
    """ Makes thumbnails and gets metadata for original file.
    """
    # Get metadata for original file.
    codec, fps, num_frames, width, height = get_metadata(video_path)

    # Create thumbnail.
    cmd = ["ffmpeg", "-y", "-i", video_path, "-vf", "scale=256:-1", "-vframes", "1", thumb_path]
    subprocess.run(cmd, check=True)

    # Create gif thumbnail.
    cmd = [
        "ffmpeg", "-y", "-r", f"{num_frames / 3}", "-i", video_path, "-vf",
        'fps=3,scale=256:-1:flags=lanczos,split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse',
        "-loop", "0", thumb_gif_path,
    ]
    subprocess.run(cmd, check=True)

    # Upload thumbnail and thumbnail gif.
    thumbnail_url = upload_file(thumb_path, host)
    thumbnail_gif_url = upload_file(thumb_gif_path, host)

    # Update the media object.
    api = get_api(host, token)
    response = api.update_media(media_id, media_update={
        'thumbnail_url': thumbnail_url,
        'thumbnail_gif_url': thumbnail_gif_url,
        'num_frames': num_frames,
        'fps': fps,
        'codec': codec,
        'width': width,
        'height': height,
    })
    logger.info(f'Thumbnail upload done! {response.message}')

if __name__ == '__main__':
    args = parse_args()
    make_thumbnails(args.host, args.token, args.media, args.input, args.output, args.gif)
